from autoplay.executors.linux import Linux

import venv


class Virtualenv(Linux):
    def proc_init(self):
        env = venv.EnvBuilder(
            system_site_packages=self.environment.pop(
                'venv_system_site_packages', False),
            clear=self.environment.pop(
                'venv_clear', False),
            symlinks=self.environment.pop(
                'venv_symlinks', False),
            upgrade=self.environment.pop(
                'venv_upgrade', True),
            with_pip=self.environment.pop(
                'venv_with_pip', True),
        )

        path = self.environment.pop('venv_path', 'venv')
        env.create(path)
        self.send(f'export PATH={path}/bin:$PATH')
        super().proc_init()



import os
import pty

from autoplay.executor import Executor

from processcontroller import ProcessController


class Linux(Executor):
    def __init__(self, schema, **options):
        super().__init__(schema, **options)
        self.exit_status = 0
        self.shell = '/bin/bash -eu'

    def wait(self):
        if self.mode == 'dryrun':
            return True
        pid, retcode = self.proc.wait()
        if self.exit_status:
            return self.exit_status
        else:
            return retcode

    def clean(self):
        return self.proc.close()

    @property
    def return_value(self):
        return self.proc.return_value

    @property
    def proc(self):
        proc = getattr(self, '_proc', None)
        if not proc:
            proc = self._proc = ProcessController()
            proc.run(self.shell.split(' '), {
                'detached': True,
                'private': True,
                'echo': self.mode != 'dryrun',
                'decode': False,
                'when': [
                    ['^AUTOPLAY_JOB_COMPLETE_TOKEN$', self.next_job],
                    ['(?!^([0-9]* )?AUTOPLAY_.*_TOKEN$)', self.print_line],
                    ['^AUTOPLAY_DONE_TOKEN$', self.next_cmd],
                    ['^.*AUTOPLAY_ERR_TOKEN$', self.abort],
                ]
            })

        return proc

    def init(self):
        for key, value in self.environment.items():
            self.send(f'export {key}="{value}"')

        for key, value in self.schema.environment.items():
            if key in self.environment:
                continue
            self.send(f'export {key}="{value}"')

        for key, value in self.doc.get('env', {}).items():
            self.send(f'export {key}="{value}"')

    def send(self, line, linetoprint=None):
        if self.mode == 'dryrun':
            print(linetoprint or line)

        else:
            if linetoprint is not None:
                if linetoprint != '':
                    print(linetoprint)
                self.proc.options['echo'] = False

            self.proc.send(line)

            if linetoprint is not None:
                self.proc.options['echo'] = True

    def dryrun(self):
        for job in self.jobs:
            for cmd in job:
                print(cmd)

    def run(self):
        if self.command_count < len(self.job):
            command = self.job[self.command_count]
            self.send('((' + command + ')' +
                      ' && echo AUTOPLAY_DONE_TOKEN )' +
                      ' || echo $? AUTOPLAY_ERR_TOKEN',
                      linetoprint=command
                      )
        else:
            if self.exit_status is None:
                self.exit_status = 0
            self.send('echo AUTOPLAY_JOB_COMPLETE_TOKEN', '')
        print()
        return self.exit_status

    @property
    def doc(self):
        if self.job_count < len(self.job_names):
            return self.schema[self.job_names[self.job_count]]
        else:
            return {}

    def next_job(self, c, l):
        self.command_count = 0
        self.job_count += 1
        for key, value in self.doc.get('env', {}).items():
            self.send(f'export {key}="{value}"')
        self()

    def next_cmd(self, c, l):
        self.command_count += 1
        self()

    def abort(self, c, l):
        self.command_count = len(self.job)
        self.job_count = len(self.jobs)
        retcode, token = l.split(b' ')
        self.exit_status = int(retcode)
        self.proc.return_value = (self.proc.pid, self.exit_status)
        self()

    def print_line(self, c, l):
        os.write(pty.STDOUT_FILENO, l)



import pprint
import sys

from autoplay.schema import Schema
from autoplay.executor import get_executor

import clilabs
import clilabs.builtins


help = clilabs.builtins.help


def main(*args):
    '''
    Bash orchestrator for yaml.

    Lets you run a job or dryrun it, or do just the setup/script/clean stage of
    a job.

    Autoplay usage:

        autoplay
        autoplay describe [job]
        autoplay run [--dryrun] [job]
        autoplay setup [--dryrun] [job]
        autoplay script [--dryrun] [job]
        autoplay clean [--dryrun] [job]
    '''
    args, kwargs = clilabs.expand(*(args or sys.argv[:]))

    if len(args) > 2:
        jobs = args[2]
        command = args[1]
    elif len(args) == 2:
        jobs = args[1]
        command = 'run'
    else:
        command = 'help'

    schema = Schema.cli()

    if command == 'help':
        print('# Found jobs:')
        print('')
        for i in schema.keys():
            print(' -', i)
        print('')
        print('# Run autoplay describe [job] to see one of them')
        print('')
        print('# Generic help:')
        return clilabs.builtins.help('autoplay.cli')

    elif command == 'describe':
        for job in jobs.split(',') if jobs else []:
            pprint.pprint(schema[job])
        return

    elif command in ['setup', 'script', 'clean']:
        kwargs['stages'] = command

    strategy = get_executor(
        kwargs.pop('executor', 'linux')
    )(schema, **kwargs)

    for name in jobs.split(','):
        strategy.load_job(name)

    strategy()
    return strategy.wait()



import pkg_resources


def get_executor(executor_name):
    executors = pkg_resources.iter_entry_points('autoplay_executors')
    for entry in executors:
        if entry.name == executor_name:
            return entry.load()


class Executor:
    modes = ['run', 'dryrun']

    def __init__(self, schema, **options):
        self.command_count = 0
        self.job_count = 0
        self.environment = options
        self.exit_status = None
        self.mode = options.pop('mode', 'run')
        self.schema = schema
        self.stages = options.pop('stages', 'setup,script,clean').split(',')
        self.environment = options
        self.jobs = []
        self.job_names = []

    def load_job(self, name):
        self.job_names.append(name)
        self.jobs.append(self.get_commands(name))

    def get_commands(self, name):
        deps = self.schema[name].get('requires', [])
        dst_cmds = []
        for stage in self.stages:
            src_cmds = self.schema[name].get(stage, [])

            if not isinstance(src_cmds, list):
                src_cmds = [src_cmds]

            for dep in deps:
                deps_cmds = self.schema[dep].get(stage, [])
                if not isinstance(deps_cmds, list):
                    deps_cmds = [deps_cmds]
                src_cmds = deps_cmds + src_cmds

            for cmd in src_cmds:
                dst_cmds.append(cmd)

        return dst_cmds

    def __call__(self):
        if not getattr(self, '_init', False):
            self.init()
            self._init = True

        if self.job_count < len(self.jobs):
            self.job = self.jobs[self.job_count]
            getattr(self, self.mode)()
        else:
            self.clean()
        return self.return_value

    def wait(self):
        return True



from setuptools import setup

setup(
    name='autoplay',
    versioning='distance',
    url='https://yourlabs.io/oss/autoplay',
    setup_requires='setupmeta',
    keywords='automation cli',
    entry_points={
        'console_scripts': [
            'autoplay = autoplay.cli:main',
        ],
        'autoplay_executors': [
            'linux = autoplay.executors.linux:Linux',
            'docker = autoplay.executors.docker:Docker',
            'venv = autoplay.executors.venv:Virtualenv',
        ],
    },
    install_requires=['clilabs', 'processcontroller'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Plugins',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Database',
        'Topic :: Software Development',
        'Topic :: System',
        'Topic :: Terminals',
    ],
    python_requires='>=3',
)



from autoplay.executor import Executor

import docker


class Docker(Executor):
    def init(self):
        self.client = docker.from_env()
        self.container = self.client.containers.run(
            self.options.get('docker_image', 'alpine'),
            detach=True,
        )

    def run(self):
        self.return_value = 0



import os
import yaml


class Schema(dict):
    def parse(self, path):
        self.shells = dict()
        self.environment = dict()
        self.plugins = dict()

        with open(path, 'r') as f:
            docs = [yaml.load(i) for i in f.read().split('---')]

        for doc in docs:
            if not doc:
                continue

            env = doc.get('env', {})
            if 'name' not in doc:
                for key, value in env.items():
                    self.environment.setdefault(key, value)
            else:
                self[doc['name']] = doc

    @classmethod
    def cli(cls):
        paths = [
            name
            for name in (
                'autoplay.yml',
                os.getenv('AUTOPLAY_PATH'),
                os.path.join(os.path.dirname(__file__), 'autoplay.yml'),
            )
            if name and os.path.exists(name)
        ]
        self = cls()
        for path in paths:
            self.parse(path)
        return self
